/*
 * authusekey - decode a key from ascii and use it
 */
#include <stdio.h>
#include <ctype.h>

#include "ntp_types.h"
#include "ntp_string.h"
#include "ntp_stdlib.h"
#include "ntp.h"

/*
 * Types of ascii representations for keys.  "Standard" means a 64 bit
 * hex number in NBS format, i.e. with the low order bit of each byte
 * a parity bit.  "NTP" means a 64 bit key in NTP format, with the
 * high order bit of each byte a parity bit.  "Ascii" means a 1-to-8
 * character string whose ascii representation is used as the key.
 */
int
authusekey(
	keyid_t keyno,
	int keytype,
	const char *str
	)
{
	int len;
	u_char key[20];

	len = authdecodekey(str, key, sizeof(key));
	if (!len)
		return 0;

	MD5auth_setkey(keyno, keytype, key, len);
	return 1;
}

/*
 * authdecodekey - decode binary or ASCII key from string
 *
 * Returns the length of the parsed key, zero if invalid.
 */
int
authdecodekey(
	const char *str,
	u_char *key,
	int max_length
	)
{
	int len;

	/*
	 * If the string is longer than 20 characters, it is
	 * a binary string encoded in hex; otherwise, it is
	 * a text string of printable ASCII characters.
	 */
	len = strlen(str);

	if (len <= 20) {
		len = min(len, max_length);
		memcpy(key, str, len);
	} else {
		char	hex[] = "0123456789abcdef";
		u_char	temp;
		char	*ptr;
		int	j;

		len = min(len / 2, max_length);
		for (j = 0; j < len * 2; j++) {
			ptr = strchr(hex, tolower(str[j]));
			if (ptr == NULL)
				return 0;
			temp = (u_char)(ptr - hex);
			if (j & 1)
				key[j / 2] |= temp;
			else
				key[j / 2] = temp << 4;
		}
	}

	return len;
}
